/** @license 2022 Neil Edelman, distributed under the terms of the
 [MIT License](https://opensource.org/licenses/MIT).

 Scan journal entries for kjv references. */

#include "../src/journal.h"
#include "../src/kjv.h"
#include "../src/helper.h"
#include <inttypes.h> /* C99 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <limits.h>

/*!conditions:re2c*/

static int scan(union date32 date, const char *const buffer,
	struct kjv *const kj) {
	const char *YYCURSOR, *YYMARKER, *yyt1, *yyt2, *yyt3,
		*s0, *s1, *t0, *t1, *line_start, *line_end;
	enum kjv_book book = Revelation;
	uint32_t chapter = 0, verse = 0, verse_end = 0;
	enum YYCONDTYPE condition = yycline;
	size_t line = 1;
	char datestr[12] = {0};
	int is_found = 0;
	assert(buffer && kj);
	YYCURSOR = YYMARKER = yyt1 = line_start = buffer;
	/*!re2c /**/
	re2c:define:YYCTYPE = char;
	re2c:yyfill:enable = 0;
	re2c:define:YYGETCONDITION = "condition";
	re2c:define:YYSETCONDITION = "condition = @@;";
	re2c:define:YYGETCONDITION:naked = 1;
	re2c:define:YYSETCONDITION:naked = 1;

	unix_control = [\x01-\x08\x0a-\x1f\x7f];
	ws = [ \t];
	glyph = [^] \ ("\x00" | "\n" | unix_control | ws);
	natural = [1-9][0-9]*;
	engage = ws+ "--" ws+;
	lookat = ws* natural ":" natural [ab]?
		("-" (natural ":")? natural [ab]?)? engage;
	first = ("I" | "1") " "?;
	second = ("II" | "2") " "?;
	third = ("III" | "3") " "?;
	*/
	for( ; ; ) { /*!re2c /**/
		<skip, book> * { goto catch; }
		<line> "\x00" { return 1; }
		<line> "\n" @line_start
			{ fprintf(stderr, "%4zu\n", line); line++; continue; }
		<line> * :=> skip /* Guess it can't be simplified? */
		<line> "Genesis" / lookat => book { book = Genesis; continue; }
		<line> "Exodus" / lookat => book { book = Exodus; continue; }
		<line> "Leviticus" / lookat => book { book = Leviticus; continue; }
		<line> "Numbers" / lookat => book { book = Numbers; continue; }
		<line> "Deuteronomy" / lookat => book { book = Deuteronomy; continue; }
		<line> "Joshua" / lookat => book { book = Joshua; continue; }
		<line> "Judges" / lookat => book { book = Judges; continue; }
		<line> "Ruth" / lookat => book { book = Ruth; continue; }
		<line> first "Samuel" / lookat => book { book = ISamuel; continue; }
		<line> second "Samuel" / lookat => book { book = IISamuel; continue; }
		<line> first "Kings" / lookat => book { book = IKings; continue; }
		<line> second "Kings" / lookat => book { book = IIKings; continue; }
		<line> first "Chronicles" / lookat
			=> book { book = IChronicles; continue; }
		<line> second "Chronicles" / lookat
			=> book { book = IIChronicles; continue; }
		<line> "Ezra" / lookat => book { book = Ezra; continue; }
		<line> "Nehemiah" / lookat => book { book = Nehemiah; continue; }
		<line> "Esther" / lookat => book { book = Esther; continue; }
		<line> "Job" / lookat => book { book = Job; continue; }
		<line> "Psalms" / lookat => book { book = Psalms; continue; }
		<line> "Proverbs" / lookat => book { book = Proverbs; continue; }
		<line> "Ecclesiastes" / lookat
			=> book { book = Ecclesiastes; continue; }
		<line> "Song of Solomon" / lookat
			=> book { book = Song_of_Solomon; continue; }
		<line> "Isaiah" / lookat => book { book = Isaiah; continue; }
		<line> "Jeremiah" / lookat => book { book = Jeremiah; continue; }
		<line> "Lamentations" / lookat
			=> book { book = Lamentations; continue; }
		<line> "Ezekiel" / lookat => book { book = Ezekiel; continue; }
		<line> "Daniel" / lookat => book { book = Daniel; continue; }
		<line> "Hosea" / lookat => book { book = Hosea; continue; }
		<line> "Joel" / lookat => book { book = Joel; continue; }
		<line> "Amos" / lookat => book { book = Amos; continue; }
		<line> "Obadiah" / lookat => book { book = Obadiah; continue; }
		<line> "Jonah" / lookat => book { book = Jonah; continue; }
		<line> "Micah" / lookat => book { book = Micah; continue; }
		<line> "Nahum" / lookat => book { book = Nahum; continue; }
		<line> "Habakkuk" / lookat => book { book = Habakkuk; continue; }
		<line> "Zephaniah" / lookat => book { book = Zephaniah; continue; }
		<line> "Haggai" / lookat => book { book = Haggai; continue; }
		<line> "Zechariah" / lookat => book { book = Zechariah; continue; }
		<line> "Malachi" / lookat => book { book = Malachi; continue; }
		<line> "Matthew" / lookat => book { book = Matthew; continue; }
		<line> "Mark" / lookat => book { book = Mark; continue; }
		<line> "Luke" / lookat => book { book = Luke; continue; }
		<line> "John" / lookat => book { book = John; continue; }
		<line> "Acts" / lookat => book { book = Acts; continue; }
		<line> "Romans" / lookat => book { book = Romans; continue; }
		<line> first "Corinthians" / lookat
			=> book { book = ICorinthians; continue; }
		<line> second "Corinthians" / lookat
			=> book { book = IICorinthians; continue; }
		<line> "Galatians" / lookat => book { book = Galatians; continue; }
		<line> "Ephesians" / lookat => book { book = Ephesians; continue; }
		<line> "Philippians" / lookat => book { book = Philippians; continue; }
		<line> "Colossians" / lookat => book { book = Colossians; continue; }
		<line> first "Thessalonians" / lookat
			=> book { book = IThessalonians; continue; }
		<line> second "Thessalonians" / lookat
			=> book { book = IIThessalonians; continue; }
		<line> first "Timothy" / lookat => book { book = ITimothy; continue; }
		<line> second "Timothy" / lookat => book { book = IITimothy; continue; }
		<line> "Titus" / lookat => book { book = Titus; continue; }
		<line> "Philemon" / lookat => book { book = Philemon; continue; }
		<line> "Hebrews" / lookat => book { book = Hebrews; continue; }
		<line> "James" / lookat => book { book = James; continue; }
		<line> first "Peter" / lookat => book { book = IPeter; continue; }
		<line> second "Peter" / lookat => book { book = IIPeter; continue; }
		<line> first "John" / lookat => book { book = IJohn; continue; }
		<line> second "John" / lookat => book { book = IIJohn; continue; }
		<line> third "John" / lookat => book { book = IIIJohn; continue; }
		<line> "Jude" / lookat => book { book = Jude; continue; }
		<line> "Revelation" / lookat => book { book = Revelation; continue; }
		<book> ws+ @s0 natural @s1 ":" @t0 natural @t1 [ab]? {
			if(chapter || verse || verse_end
				|| !helper_natural(s0, s1, &chapter)
				|| !helper_natural(t0, t1, &verse)) goto catch;
			continue;
		}
		<book> "-" @s0 natural @s1 [ab]? { /* Verse range. */
			if(!chapter || !verse || verse_end
				|| !helper_natural(s0, s1, &verse_end)) goto catch;
			continue;
		}
		<book> engage => skip {
			const size_t old_set_words = kj->set_words;
			char citestr[12];
			if(!chapter || !verse || verse_end && verse_end <= verse)
				goto catch;
			union kjvcite cite
				= { .book = book, .chapter = chapter, .verse = verse };
			if(!datestr[0]) date32_to_string(date, &datestr); /* Only once. */
			kjvcite_to_string(cite, &citestr);
			for( ; ; verse++, cite.verse++) {
				if(!kjv_add(kj, cite)) goto catch;
				if(!verse_end || verse_end <= verse) break;
			}
			if(verse_end) {
				printf("%s\t%zu\t%zu\t# %s-%" PRIu32 "\n",
					datestr, old_set_words, kj->set_words, citestr, verse_end);
			} else {
				printf("%s\t%zu\t%zu\t# %s\n",
					datestr, old_set_words, kj->set_words, citestr);
			}
			book = Revelation, chapter = 0, verse = 0, verse_end = 0;
			is_found = 1;
			continue;
		}
		<skip> [^\n\x00]* @line_end "\n" @s0 => line {
			const size_t size = (size_t)(line_end - line_start);
			int intsize = size > 40 ? 40 : (int)size;
			fprintf(stderr, "%s%4zu: %.*s%s\n", is_found ? "\033[1;35m" : "",
				line, intsize, line_start, is_found ? "\033[0m" : "");
			line_start = s0, is_found = 0;
			line++;
			continue;
		}
	*/ }
catch:
	if(!errno) errno = EILSEQ;
	{
		char a[12];
		date32_to_string(date, &a);
		fprintf(stderr, "%s line %zu: unexpected.\n", a, line);
	}
	return 0;
}

int main(void) {
	int success = EXIT_SUCCESS;
	struct journal j;
	struct journal_iterator it;
	struct kjv kj = kjv();
	union date32 k;
	union load *v;
	size_t i;
	/*scan((union date32){.year=2000, .month=1, .day=1}, "\n\n\n"
		 "Romans 3:23 -- ``For all have sinned, "
		 "and come short of the glory of God.''\n", &kj);*/
	j = journal();
	if(!journal_is_valid(&j)) goto catch;
	fprintf(stderr, "Journal: %s.\n", journal_to_string(&j));
	printf("set term postscript eps enhanced\n"
		"set output \"kjv.eps\"\n"
		"$Data <<EOD\n"
		"# date\told\tnew / %zu\n", kj.total_words);
	it = journal_begin(&j), i = 0; while(journal_next(&it, &k, &v)) {
		if(!scan(k, v->text, &kj)) goto catch;
		if(++i > 32) break;
	}
	printf("EOD\n"
		"set monochrome\n"
		"set xdata time\n"
		"set timefmt \"%%Y-%%m-%%d\"\n"
		"set xtics format \"%%Y-%%m-%%d\" rotate by -30\n"
		"set ylabel \"KJV memorized\"\n"
		"set format y \"%%g%%%%\"\n"
		"unset key #set key bottom right\n"
		"set grid\n"
		"unset border\n"
		"plot $Data using 1:($3-$2)*100/%zu smooth cumulative "
		"with steps lw 2\n", kj.total_words);
	goto finally;
catch:
	success = EXIT_FAILURE;
	perror("journal");
finally:
	journal_(&j);
	return success;
}
